"""Air uses custom response classes to improve the developer experience."""

from collections.abc import Mapping
from typing import override

from fastapi import status
from starlette.background import BackgroundTask
from starlette.datastructures import URL
from starlette.responses import (
    FileResponse as FileResponse,
    HTMLResponse as HTMLResponse,
    JSONResponse as JSONResponse,
    PlainTextResponse as PlainTextResponse,
    RedirectResponse as StarletteRedirectResponse,
    Response as Response,
    StreamingResponse as StreamingResponse,
)
from starlette.types import Send

from .tags import BaseTag


class AirResponse(HTMLResponse):
    """Response class to handle air.tags.Tags or HTML (from Jinja2)."""

    @override
    def render(self, tag: BaseTag | str) -> bytes:
        """Render Tag elements to bytes of HTML."""
        return super().render(str(tag))


TagResponse = AirResponse  # Alias for clarity


class SSEResponse(StreamingResponse):
    """Response class for Server Sent Events

    Example:

        # For tags
        import random
        from asyncio import sleep

        import air

        app = air.Air()


        @app.page
        def index():
            return air.layouts.mvpcss(
                air.Script(src="https://unpkg.com/htmx-ext-sse@2.2.1/sse.js"),
                air.Title("Server Sent Event Demo"),
                air.H1("Server Sent Event Demo"),
                air.P("Lottery number generator"),
                air.Section(
                    hx_ext="sse",
                    sse_connect="/lottery-numbers",
                    hx_swap="beforeend show:bottom",
                    sse_swap="message",
                ),
            )

        async def lottery_generator():
            while True:
                lottery_numbers = ", ".join([str(random.randint(1, 40)) for x in range(6)])
                # Tags work seamlessly
                yield air.Aside(lottery_numbers)
                # As do strings. Non-strings are cast to strings via the str built-in
                yield "Hello, world"
                await sleep(1)


        @app.get("/lottery-numbers")
        async def get():
            return air.SSEResponse(lottery_generator())
    """

    media_type = "text/event-stream"

    async def stream_response(self, send: Send) -> None:
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            },
        )
        async for chunk in self.body_iterator:
            if not isinstance(chunk, bytes | memoryview):
                lines = list(str(chunk).splitlines())
                formatted = [f"data: {t}" for t in lines]
                data = "\n".join(formatted)
                chunk = f"event: message\n{data}\n\n"
                chunk = chunk.encode(self.charset)
            await send({"type": "http.response.body", "body": chunk, "more_body": True})

        await send({"type": "http.response.body", "body": b"", "more_body": False})


class RedirectResponse(StarletteRedirectResponse):
    def __init__(
        self,
        url: str | URL,
        status_code: int = status.HTTP_307_TEMPORARY_REDIRECT,
        headers: Mapping[str, str] | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        super().__init__(url=url, status_code=status_code, headers=headers, background=background)



"""Display and Validation of HTML forms. Powered by pydantic.

Pro-tip: Always validate incoming data."""

import re
from collections.abc import Callable, Sequence
from types import UnionType
from typing import Any, Literal, Self, Union, get_args, get_origin

import annotated_types
from pydantic import AliasChoices, AliasPath, BaseModel, Discriminator, Field, JsonValue, ValidationError
from pydantic.fields import Deprecated, FieldInfo
from pydantic_core import ErrorDetails, PydanticUndefined
from starlette.datastructures import FormData

from . import tags
from .requests import Request
from .tags import SafeStr


class AirForm:
    """
    A form handler that validates incoming form data against a Pydantic model. Can be used with awaited form data or with FastAPI's dependency injection system.

    Example:

        class FlightModel(BaseModel):
            flight_number: str
            destination: str

        class FlightForm(air.AirForm):
            model = FlightModel

        @app.post("/flight")
        async def flight_form(request: air.Request):
            "Awaited form data"
            flight = await FlightForm.from_request(request)
            if flight.is_valid:
                return air.Html(air.H1(flight.data.flight_number))
            return air.Html(air.H1(air.Raw(str(len(flight.errors)))))

        @app.post("/flight-depends")
        async def flight_form_depends(flight: Annotated[FlightForm, Depends(FlightForm())]):
            "Dependency injection"
            if flight.is_valid:
                return air.Html(air.H1(flight.data.flight_number))
            return air.Html(air.H1(air.Raw(str(len(flight.errors)))))


    NOTE: This is named AirForm to avoid collisions with tags.Form
    """

    model: type[BaseModel] | None = None
    data: Any = None  # TODO change type to something more specific
    initial_data: dict | None = None
    errors: list[ErrorDetails] | None = None
    is_valid: bool = False
    includes: Sequence[str] | None = None

    def __init__(self, initial_data: dict | None = None) -> None:
        if self.model is None:
            msg = "model"
            raise NotImplementedError(msg)
        self.initial_data = initial_data

    async def __call__(self, form_data: dict[Any, Any] | FormData) -> Self:
        self.validate(form_data)
        return self

    def validate(self, form_data: dict[Any, Any] | FormData) -> bool:
        # Store the submitted data to preserve values on error
        self.submitted_data = dict(form_data) if hasattr(form_data, "items") else form_data
        try:
            self.data = self.model(**form_data)
            self.is_valid = True
        except ValidationError as e:
            self.errors = e.errors()
        return self.is_valid

    @classmethod
    async def from_request(cls, request: Request) -> Self:
        form_data = await request.form()
        self = cls()
        await self(form_data=form_data)
        return self

    @property
    def widget(self) -> Callable:
        """Widget for rendering of form in HTML

        If you want a custom widget, replace with a function that accepts:

            - model: BaseModel
            - data: dict|None
            - errors:dict|None=None
        """
        return default_form_widget

    def render(self) -> SafeStr:
        # Use submitted data if available (preserves values on validation errors)
        render_data = getattr(self, "submitted_data", None) or self.initial_data
        return SafeStr(
            self.widget(
                model=self.model,
                data=render_data,
                errors=self.errors,
                includes=self.includes,
            ),
        )


def pydantic_type_to_html_type(field_info: Any) -> str:
    """Return HTML type from pydantic type.

    Default to 'text' for unknown types.
    """
    special_fields = [
        "hidden",
        "email",
        "password",
        "url",
        "datedatetime-local",
        "month",
        "time",
        "color",
        "file",
    ]
    for field in special_fields:
        if field_info.json_schema_extra and field_info.json_schema_extra.get(field, False):
            return field

    return {int: "number", float: "number", bool: "checkbox", str: "text"}.get(field_info.annotation, "text")


def get_user_error_message(error: dict) -> str:
    """Convert technical pydantic error to user-friendly message."""
    error_type = error.get("type", "")
    technical_msg = error.get("msg", "")

    # Map error types to user-friendly messages
    messages = {
        "missing": "This field is required.",
        "int_parsing": "Please enter a valid number.",
        "float_parsing": "Please enter a valid number.",
        "bool_parsing": "Please select a valid option.",
        "string_too_short": "This value is too short.",
        "string_too_long": "This value is too long.",
        "value_error": "This value is not valid.",
        "type_error": "Please enter the correct type of value.",
        "assertion_error": "This value doesn't meet the requirements.",
        "url_parsing": "Please enter a valid URL.",
        "email": "Please enter a valid email address.",
        "json_invalid": "Please enter valid JSON.",
        "enum": "Please select a valid option.",
        "greater_than": "This value must be greater than the minimum.",
        "greater_than_equal": "This value must be at least the minimum.",
        "less_than": "This value must be less than the maximum.",
        "less_than_equal": "This value must be at most the maximum.",
    }

    # Get user-friendly message or fallback to technical message
    return messages.get(error_type, technical_msg or "Please correct this error.")


def errors_to_dict(errors: list[dict] | None) -> dict[str, dict]:
    """Converts a pydantic error list to a dictionary for easier reference."""
    if errors is None:
        return {}
    return {error["loc"][0]: error for error in errors}


def default_form_widget(
    model: type[BaseModel],
    data: dict | None = None,
    errors: list | None = None,
    includes: Sequence[str] | None = None,
) -> str:
    """
    Render a form widget for a given Pydantic model.

    Args:
        model: The Pydantic model class to render
        data: Dictionary of data to pre-populate
        errors: List of Pydantic validation errors
        includes: Sequence of field names to include (None means all)

    Returns:
        HTML string representing the form
    """
    error_dict = errors_to_dict(errors)
    fields = []
    for field_name, field_info in model.model_fields.items():
        if includes is not None and field_name not in includes:
            continue
        field_type = field_info.annotation
        origin = get_origin(field_type)

        # Handle optional types (Union with None)
        if (origin is Union or origin is UnionType) and type(None) in get_args(field_type):
            # This is a Union type, get the non-None type
            args = get_args(field_type)
            field_type = next((arg for arg in args if arg is not type(None)), str)
        input_type = pydantic_type_to_html_type(field_info)
        kwargs = {}
        # Inject values
        if data is not None and field_name in data:
            kwargs["value"] = data[field_name]

        error = error_dict.get(field_name)
        if error:
            kwargs["aria-invalid"] = "true"
        json_schema_extra: dict = field_info.json_schema_extra or {}
        if json_schema_extra.get("autofocus"):
            kwargs["autofocus"] = True
        fields.append(
            tags.Tags(
                tags.Label(
                    json_schema_extra.get("label") or field_name,
                    for_=field_name,
                ),
                tags.Input(name=field_name, type=input_type, id=field_name, **kwargs),
                (tags.Small(get_user_error_message(error), id=f"{field_name}-error") if error else ""),
            ),
        )

    return tags.Tags(*fields).render()


def AirField(
    default: Any = PydanticUndefined,
    *,
    default_factory: Callable[[], Any] | Callable[[dict[str, Any]], Any] | None = None,
    alias: str | None = None,
    alias_priority: int | None = None,
    validation_alias: str | AliasPath | AliasChoices | None = None,
    serialization_alias: str | None = None,
    title: str | None = None,
    field_title_generator: Callable[[str, FieldInfo], str] | None = None,
    description: str | None = None,
    examples: list[Any] | None = None,
    exclude: bool | None = None,
    exclude_if: Callable[[Any], bool] | None = None,
    discriminator: str | Discriminator | None = None,
    deprecated: Deprecated | str | bool | None = None,
    json_schema_extra: dict[str, JsonValue] | None = None,
    frozen: bool | None = None,
    validate_default: bool | None = None,
    repr: bool | None = None,
    init: bool | None = None,
    init_var: bool | None = None,
    kw_only: bool | None = None,
    pattern: str | re.Pattern[str] | None = None,
    strict: bool | None = None,
    coerce_numbers_to_str: bool | None = None,
    gt: annotated_types.SupportsGt | None = None,
    ge: annotated_types.SupportsGe | None = None,
    lt: annotated_types.SupportsLt | None = None,
    le: annotated_types.SupportsLe | None = None,
    multiple_of: float | None = None,
    allow_inf_nan: bool | None = None,
    max_digits: int | None = None,
    decimal_places: int | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    union_mode: Literal["smart", "left_to_right"] | None = None,
    fail_fast: bool | None = None,
    # Not in pydantic.Field:
    type: str | None = None,
    label: str | None = None,
    autofocus: bool = False,
    **extra: Any,
) -> Any:
    """A wrapper around pydantic.Field to provide a cleaner interface for defining
    special input types and labels in air forms.

    NOTE: This is named AirField to adhere to the same naming convention as AirForm.

    Example:

        class CheeseModel(BaseModel):
            name: str = air.AirField(label="Name", autofocus=True)
            age: int

        class CheeseForm(air.AirForm):
            model = CheeseModel

    Used with FastAPI's dependency injection system:
        class CheeseModel(pydantic.BaseModel):
            name: str
            age: int

        class CheeseForm(air.AirForm):
            model = CheeseModel

        @app.post("/cheese")
        async def cheese_form(cheese: Annotated[CheeseForm, Depends(CheeseForm())]):
            if cheese.is_valid:
                return air.Html(air.H1(cheese.data.name))
            return air.Html(air.H1(air.Raw(str(len(cheese.errors)))))
    """
    schema_extra: dict[str, JsonValue] = json_schema_extra or {}
    if type:
        schema_extra[type] = True
    if label:
        schema_extra["label"] = label
    if autofocus:
        schema_extra["autofocus"] = True
    if extra:
        schema_extra.update(extra)

    # noinspection PyArgumentList
    return Field(
        default,
        default_factory=default_factory,
        alias=alias,
        alias_priority=alias_priority,
        validation_alias=validation_alias,
        serialization_alias=serialization_alias,
        title=title,
        field_title_generator=field_title_generator,
        description=description,
        examples=examples,
        exclude=exclude,
        exclude_if=exclude_if,
        discriminator=discriminator,
        deprecated=deprecated,
        json_schema_extra=schema_extra,
        frozen=frozen,
        validate_default=validate_default,
        repr=repr,
        init=init,
        init_var=init_var,
        kw_only=kw_only,
        pattern=pattern,
        strict=strict,
        coerce_numbers_to_str=coerce_numbers_to_str,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        allow_inf_nan=allow_inf_nan,
        max_digits=max_digits,
        decimal_places=decimal_places,
        min_length=min_length,
        max_length=max_length,
        union_mode=union_mode,
        fail_fast=fail_fast,
    )  # ty: ignore[no-matching-overload]



"""Utilities for the Air Tag system."""

from __future__ import annotations

import base64
import html
import tempfile
import webbrowser
from io import StringIO
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final
from urllib.error import URLError

from air.exceptions import BrowserOpenError

type StrPath = PathLike | Path | str

if TYPE_CHECKING:
    from rich.console import Console

EXTRA_FEATURE_PRETTY_ERROR_MESSAGE: Final = (
    "Extra feature 'pretty' is not installed. Install with: `uv add air[pretty]`"
)
FORMAT_HTML_ENCODING: Final = "unicode"
HTML_DOCTYPE: Final = "<!doctype html>"
DEFAULT_THEME: Final = "dracula"
PANEL_TITLE: Final = "Air → HTML"
PANEL_TITLE_STYLE: Final = "italic bold"
PANEL_BORDER_STYLE: Final = "bright_magenta"
SYNTAX_LEXER: Final = "html"
DATA_URL_MAX: Final[int] = 32_000
BLOB_URL_PRESET = "data:text/html;charset=utf-8;base64,"


def clean_html_attr_key(key: str) -> str:
    """Normalize attribute names to align with HTML conventions.

    Args:
        key: Attribute name supplied by the caller.

    Returns:
        The normalized attribute name compatible with HTML.
    """
    # If a "_"-suffixed proxy for "class", "for", or "id" is used,
    # convert it to its normal HTML equivalent.
    key = {"class_": "class", "for_": "for", "id_": "id", "as_": "as"}.get(key, key)
    # Remove leading underscores and replace underscores with dashes
    return key.lstrip("_").replace("_", "-")


def pretty_format_html(
    source: str,
    *,
    with_body: bool = False,
    with_head: bool = False,
    with_doctype: bool = False,
) -> str:
    """Pretty-print HTML and unescape common entities in the result.

    Args:
        source: Raw HTML markup to format.
        with_body: Whether to wrap the markup in a `<body>` element.
        with_head: Whether to include a `<head>` element when `with_body` is set.
        with_doctype: Whether to prefix the result with a doctype declaration.

    Returns:
        The formatted HTML string with entities such as `&lt;` unescaped.

    Note:
        Entity unescaping applies to attribute values as well; use this helper only with trusted HTML.
    """
    return html.unescape(
        format_html(source, with_body=with_body, with_head=with_head, with_doctype=with_doctype, pretty=True)
    )


def format_html(
    source: str,
    *,
    with_body: bool = False,
    with_head: bool = False,
    with_doctype: bool = False,
    pretty: bool = False,
) -> str:
    """Format HTML markup using `lxml`.

    Args:
        source: Raw HTML markup to format.
        with_body: Whether to ensure a `<body>` element is present.
        with_head: Whether to add a `<head>` element when `with_body` is enabled.
        with_doctype: Whether to include a doctype declaration in the result.
        pretty: Whether to indent the output for readability.

    Returns:
        The serialized HTML produced by `lxml.html.tostring`.

    Raises:
        ModuleNotFoundError: If the optional `lxml` dependency is unavailable.
    """
    try:
        from lxml import (
            etree,  # ty: ignore[unresolved-import]
            html as l_html,
        )
    except ModuleNotFoundError:
        raise ModuleNotFoundError(EXTRA_FEATURE_PRETTY_ERROR_MESSAGE) from None
    else:
        if with_body:
            source = l_html.document_fromstring(source, ensure_head_body=with_head)
        else:
            source = l_html.fromstring(source)
        if pretty:
            etree.indent(source)  # pretty indentation
        doctype = HTML_DOCTYPE if with_doctype else None
        return l_html.tostring(source, encoding=FORMAT_HTML_ENCODING, pretty_print=pretty, doctype=doctype)


def open_local_file_in_the_browser(path: StrPath) -> None:
    """Open a local HTML file in the default browser.

    Args:
        path: Path to a file or directory containing an `index.html`.

    Raises:
        FileNotFoundError: The path does not exist or `index.html` is missing.
        BrowserOpenError: The browser command failed to launch.
    """
    path = Path(path)
    if path.is_dir():
        path /= "index.html"
    if not path.exists():
        raise FileNotFoundError(path)

    url = path.expanduser().resolve().as_uri()
    _open_new_tab(url)


def _open_new_tab(url: str) -> None:
    """Launch a new browser tab for the provided URL.

    Args:
        url: The URL to open.

    Raises:
        BrowserOpenError: The browser invocation returned a failure signal.
    """
    open_new_tab_successfully = webbrowser.open_new_tab(url)
    if not open_new_tab_successfully:
        msg = f"Could not open browser for URI: {url}. "
        raise BrowserOpenError(msg)


def open_html_blob_in_the_browser(html_source: str, *, data_url_max: int = DATA_URL_MAX) -> None:
    """Open HTML content encoded as a data URL in the browser.

    Args:
        html_source: HTML markup to embed in a data URL.
        data_url_max: Maximum permitted URL length before falling back to a file.

    Raises:
        URLError: The data URL exceeds the configured maximum length.
    """
    source_bytes = html_source.encode()
    url = BLOB_URL_PRESET + base64.b64encode(source_bytes).decode("ascii")
    if len(url) >= data_url_max:
        msg = "html_source is to long!"
        raise URLError(msg)
    _open_new_tab(url)


def open_html_in_the_browser(html_source: str) -> None:
    """Open an HTML string in the browser via a temporary file.

    Args:
        html_source: HTML markup to render in the browser.

    Raises:
        BrowserOpenError: The browser command failed.
    """
    with tempfile.NamedTemporaryFile("w", delete=False, suffix=".html", encoding="utf-8") as f:
        f.write(html_source)
        path = Path(f.name)

    _open_new_tab(path.as_uri())


def save_pretty_html(
    source: str,
    *,
    theme: str = DEFAULT_THEME,
    file_path: StrPath,
) -> None:
    """Persist syntax-highlighted HTML to a file.

    Args:
        source: HTML markup to render with syntax highlighting.
        theme: Rich syntax highlighting theme name.
        file_path: Destination file path for the exported HTML.
    """
    console = _get_pretty_html_console(source, theme=theme, record=True)
    console.save_html(path=str(file_path))


def display_pretty_html_in_the_browser(
    source: str,
    *,
    theme: str = DEFAULT_THEME,
) -> None:
    """Open syntax-highlighted HTML in the browser.

    Args:
        source: HTML markup to render with syntax highlighting.
        theme: Rich syntax highlighting theme name.
    """
    open_html_in_the_browser(export_pretty_html(source, theme=theme))


def export_pretty_html(
    source: str,
    *,
    theme: str = DEFAULT_THEME,
) -> str:
    """Return syntax-highlighted HTML for display elsewhere.

    Args:
        source: HTML markup to render with syntax highlighting.
        theme: Rich syntax highlighting theme name.

    Returns:
        The rendered HTML containing the highlighted markup.
    """
    console = _get_pretty_html_console(source, theme=theme, record=True)
    return console.export_html()


def pretty_print_html(
    source: str,
    *,
    theme: str = DEFAULT_THEME,
    record: bool = False,
) -> None:
    """Render HTML with syntax highlighting inside a styled terminal panel.

    Args:
        source: HTML markup to render.
        theme: Rich syntax highlighting theme name.
        record: Whether to buffer the output for later export.

    Raises:
        ModuleNotFoundError: The optional Rich dependency is unavailable.
    """
    _get_pretty_html_console(source, theme=theme, record=record)


def _get_pretty_html_console(
    source: str,
    *,
    theme: str = DEFAULT_THEME,
    record: bool = False,
) -> Console:
    """Return a Rich console configured for HTML syntax highlighting.

    Args:
        source: HTML markup to render.
        theme: Rich syntax highlighting theme name.
        record: Whether to buffer the console output.

    Returns:
        A configured Rich console instance.

    Raises:
        ModuleNotFoundError: The optional Rich dependency is unavailable.
    """
    try:
        from rich import box
        from rich.console import Console
        from rich.panel import Panel
        from rich.syntax import Syntax
        from rich.text import Text
    except ModuleNotFoundError:
        raise ModuleNotFoundError(EXTRA_FEATURE_PRETTY_ERROR_MESSAGE) from None
    else:
        syntax = Syntax(source, SYNTAX_LEXER, theme=theme, line_numbers=True, indent_guides=True, word_wrap=True)
        title = Text(PANEL_TITLE, style=PANEL_TITLE_STYLE)
        panel = Panel.fit(
            syntax,
            box=box.HEAVY,
            border_style=PANEL_BORDER_STYLE,
            title=title,
        )
        buffer = StringIO() if record else None
        console = Console(record=record, file=buffer)
        console.print(panel, soft_wrap=False)
        return console


def locals_cleanup(
    data: dict[str, Any],
    _skip: frozenset[str] = frozenset({"self", "children", "text_child", "args", "kwargs"}),
) -> dict[str, Any]:
    """Filter local variables for keyword argument construction.

    Args:
        data: Dictionary of local variables to filter.
        _skip: Keys that should remain excluded from the result.

    Returns:
        A dictionary containing only keyword-safe values.
    """
    return {key: value for key, value in data.items() if value is not None and key[0] != "_" and key not in _skip}


class SafeStr(str):
    """String subclass that bypasses HTML escaping when rendered."""



from __future__ import annotations

from fastapi.exceptions import HTTPException as FASTAPIHTTPException


class HTTPException(FASTAPIHTTPException):
    """Convenience import from FastAPI"""


class BaseAirException(Exception):
    """Base AIR Exception"""


class RenderException(BaseAirException):
    """Error thrown when a render function fails."""


class ObjectDoesNotExist(HTTPException):
    """Thrown when a record in a persistence store can't be found."""


class BrowserOpenError(RuntimeError):
    """Opening the browser failed."""



from collections.abc import Callable
from typing import TYPE_CHECKING, NoReturn

if TYPE_CHECKING:
    GitHubOAuthRouterFactory: Callable

try:
    from . import auth as auth
except ImportError:  # pragma: no cover
    msg = "air.ext.auth requires installing the authlib package."

    class NotImportable:
        def __getattribute__(self, name: str) -> NoReturn:
            raise RuntimeError(msg)

        def __str__(self) -> str:
            return msg

        def __repr__(self) -> str:
            return msg

    auth = NotImportable()  # ty:ignore[invalid-assignment]


try:
    from . import sqlmodel as sqlmodel
except ImportError:  # pragma: no cover
    msg = "air.ext.sqlmodel requires installing the sqlmodel and greenlet packages. "
    msg += "It may also require the installation of database specific connectors."

    class NotImportable:
        def __getattribute__(self, name: str) -> NoReturn:
            raise RuntimeError(msg)

        def __str__(self) -> str:
            return msg

        def __repr__(self) -> str:
            return msg

    sqlmodel = NotImportable()  # ty:ignore[invalid-assignment]



from collections.abc import Callable
from typing import Final

from starlette.requests import Request
from starlette.types import ASGIApp, Receive, Scope, Send

from .exceptions import HTTPException
from .layouts import mvpcss
from .responses import AirResponse
from .tags import H1, P, Title


def default_404_router_handler(router_name: str) -> ASGIApp:
    """Build an ASGI app that delegates the 404 to the default_404_exception_handler."""

    async def app(scope: Scope, receive: Receive, send: Send) -> None:
        # Create a Request so router handler can render URL and other context.
        request: Request = Request(scope, receive=receive)

        # Use a concrete HTTPException (status 404). Your handler accepts Exception.
        exc: HTTPException = HTTPException(
            status_code=404,
            detail=f"Not Found in router '{router_name}'",
        )

        # Delegate to the project's default 404 renderer (AirResponse).
        response = default_404_exception_handler(request, exc)

        # AirResponse is a Starlette-compatible Response, so call it as ASGI.
        await response(scope, receive, send)

    return app


def default_404_exception_handler(request: Request, exc: Exception) -> AirResponse:
    """Default 404 exception handler. Can be overloaded."""
    return AirResponse(
        mvpcss(
            Title("404 Not Found"),
            H1("404 Not Found"),
            P("The requested resource was not found on this server."),
            P(f"URL: {request.url}"),
        ),
        status_code=404,
    )


def default_500_exception_handler(request: Request, exc: Exception) -> AirResponse:
    """Default 500 exception handler. Can be overloaded."""
    return AirResponse(
        mvpcss(
            Title("500 Internal Server Error"),
            H1("500 Internal Server Error"),
            P("An internal server error occurred."),
        ),
        status_code=500,
    )


type ExceptionHandlersType = dict[int, Callable[[Request, Exception], AirResponse]]
DEFAULT_EXCEPTION_HANDLERS: Final[ExceptionHandlersType] = {
    404: default_404_exception_handler,
    500: default_500_exception_handler,
}



"""Tools for handling dependencies, for things like handling incoming data from client libraries like HTMX."""

from fastapi import Depends, Header


def _is_htmx_request(hx_request: str = Header(default=None)) -> bool:
    """Dependency injection function used to determine if request is coming from HTMX. Checks if 'hx-request' header is in the HTTP request.

    Returns:
        bool: Whether or not a request is coming from an HTMX action.

    Example:

        import air

        app = air.Air()


        @app.get("/")
        def index_get(is_htmx: bool = air.is_htmx_request):
            return air.H1(f"Is HTMX request?: {is_htmx}")


        @app.post("/")
        def index_post(is_htmx: bool = air.is_htmx_request):
            return air.H1(f"Is HTMX request?: {is_htmx}")


        @app.patch("/")
        def index_patch(is_htmx: bool = air.is_htmx_request):
            return air.AirResponse(air.H1(f"Is HTMX request?: {is_htmx}"))


        @app.put("/")
        def index_put(is_htmx: bool = air.is_htmx_request):
            return air.AirResponse(air.H1(f"Is HTMX request?: {is_htmx}"))


        @app.delete("/")
        def index_delete(is_htmx: bool = air.is_htmx_request):
            return air.AirResponse(air.H1(f"Is HTMX request?: {is_htmx}"))
    """
    return hx_request is not None and hx_request.lower() == "true"


is_htmx_request = Depends(_is_htmx_request)



"""Tools for building layouts and several simple layouts for quick prototyping."""

from typing import Any

from .tags import Body, Children, Head, Header, Html, Link, Main, Script, Style
from .tags.types import HEAD_TAG_TYPES, AttributesType


def filter_body_tags(tags) -> list:
    """Given a list of tags, only list the ones that belong in body of an HTML document."""
    return [t for t in tags if not isinstance(t, HEAD_TAG_TYPES)]


def filter_head_tags(tags) -> list:
    """Given a list of tags, only list the ones that belong in head of an HTML document."""
    return [t for t in tags if isinstance(t, HEAD_TAG_TYPES)]


def _header(tags) -> Header | str:
    """Extracts the air.Header tag from a set of tags."""
    for tag in tags:
        if isinstance(tag, Header):
            return tag
    return ""


def mvpcss(*children: Any, is_htmx: bool = False, **kwargs: AttributesType) -> Html | Children:
    """Renders the basic layout with MVP.css and HTMX for quick prototyping

    1. At the top level HTML head tags are put in the `<head>` tag
    2. Otherwise everything is put in the `<body>`
    3. `Header` and `Nav` tags are placed in the top of the body above the `Main` tag
    4. If `is_htmx` is True, then the layout isn't included. This is to support the `hx_boost`
        feature of HTMX

    The `mvpcss` function is a quick prototyping tool. It isn't designed to be extensible.
        Rather the `mvpcss` layout function makes it easy to roll out quick demonstrations and proofs-of-concept.
        For more advanced layouts like Eidos or a full-fledged MVP.css implementation,
        you'll have to create your own layouts.

    Args:
        children: These typically inherit from air.Tag but can be anything
        is_htmx: Whether or not HTMX sent the request from the page

    Example:

        from fastapi import Depends
        import air

        app = air.Air()


        @app.page
        async def index(is_htmx: bool = Depends(air.is_htmx_request)):
            return air.layouts.mvpcss(
                air.Title("Home"),
                air.Article(
                    air.H1("Welcome to Air"), air.P(air.A("Click to go to Dashboard", href="/dashboard")), hx_boost="true"
                ),
                is_htmx=is_htmx
            )


        @app.page
        async def dashboard(is_htmx: bool = Depends(air.is_htmx_request)):
            return air.layouts.mvpcss(
                air.Title("Dashboard"), air.Article(air.H1("Dashboard"), air.P(air.A("Go home", href="/")), hx_boost="true"),
                is_htmx=is_htmx
            )

    """
    body_tags = filter_body_tags(children)
    head_tags = filter_head_tags(children)

    if is_htmx:
        return Children(Main(*body_tags), *head_tags)

    return Html(
        Head(
            Link(rel="stylesheet", href="https://unpkg.com/mvp.css"),
            Style("footer, header, main { padding: 1rem; } nav {margin-bottom: 1rem;}"),
            Script(
                src="https://cdn.jsdelivr.net/npm/htmx.org@2.0.6/dist/htmx.min.js",
                integrity="sha384-Akqfrbj/HpNVo8k11SXBb6TlBWmXXlYQrCSqEWmyKJe+hDm3Z/B2WVG4smwBkRVm",
                crossorigin="anonymous",
            ),
            *head_tags,
        ),
        Body(
            _header(body_tags),
            Main(*[x for x in body_tags if not isinstance(x, Header)]),
        ),
    )


def picocss(*children: Any, is_htmx: bool = False, **kwargs: AttributesType) -> Html | Children:
    """Renders the basic layout with PicoCSS and HTMX for quick prototyping

    1. At the top level HTML head tags are put in the `<head>` tag
    2. Otherwise everything is put in the `<body>`
    3. If `is_htmx` is True, then the layout isn't included. This is to support the `hx_boost`
        feature of HTMX

    Note: `PicoCSS` is a quick prototyping tool. It isn't designed to be extensible.
        Rather the `picocss` layout function makes it easy to roll out quick demonstrations and proofs-of-concept.
        For more advanced layouts like Eidos or a full-fledged PicoCSS implementation,
        you'll have to create your own layouts.

    Args:
        children: These typically inherit from air.Tag but can be anything
        is_htmx: Whether or not HTMX sent the request from the page

    """
    body_tags = filter_body_tags(children)
    head_tags = filter_head_tags(children)

    if is_htmx:
        return Children(Main(*body_tags, class_="container"), *head_tags)

    return Html(
        Head(
            Link(
                rel="stylesheet",
                href="https://cdn.jsdelivr.net/npm/@picocss/pico@2/css/pico.min.css",
            ),
            Script(
                src="https://cdn.jsdelivr.net/npm/htmx.org@2.0.6/dist/htmx.min.js",
                integrity="sha384-Akqfrbj/HpNVo8k11SXBb6TlBWmXXlYQrCSqEWmyKJe+hDm3Z/B2WVG4smwBkRVm",
                crossorigin="anonymous",
            ),
            *head_tags,
        ),
        Body(Main(*body_tags, class_="container")),
    )



"""Tools for handling requests."""

import json
from dataclasses import dataclass, field
from typing import Any, Final
from urllib.parse import urlsplit, urlunsplit

from starlette.datastructures import Headers as Headers
from starlette.requests import Request as Request

# HTMX Header names as constants
HX_REQUEST: Final = "HX-Request"
HX_BOOSTED: Final = "HX-Boosted"
HX_CURRENT_URL: Final = "HX-Current-URL"
HX_HISTORY_RESTORE_REQUEST: Final = "HX-History-Restore-Request"
HX_PROMPT: Final = "HX-Prompt"
HX_TARGET: Final = "HX-Target"
HX_TRIGGER: Final = "HX-Trigger"
HX_TRIGGER_NAME: Final = "HX-Trigger-Name"

# Some servers send either of these for the triggering event payload
TRIGGERING_EVENT_ALIASES: Final[tuple[str, ...]] = ("Triggering-Event", "HX-Triggering-Event")


@dataclass(slots=True)
class HtmxDetails:
    """
    Attached to every Request served by Air; provides helpers for HTMX-aware handling.
    Derived values are computed once in `__post_init__`.
    """

    # fields
    headers: Headers
    url: str

    # Derived fields (formerly properties)
    is_hx_request: bool = field(init=False)
    boosted: bool = field(init=False)
    """`True` if the request came from an element with the `hx-boost` attribute. Detected by checking if the `HX-Boosted` header equals `true`.

    Example:

        import air
        from random import randint

        app = air.Air()


        @app.page
        def index(request: air.Request):

            if request.htmx.boosted:
                # Do something here
    """
    current_url: str | None = field(init=False)
    """The current URL in the browser that htmx made this request from, or `None` for non-htmx requests. Based on the `HX-Current-URL` header."""
    current_url_abs_path: str | None = field(init=False)
    """The absolute-path form of `current_url`, that is the URL without scheme or netloc, or None for non-htmx requests.

    This value will also be `None` if the scheme and netloc do not match the request. This could happen if the request is cross-origin, or if Air is not configured correctly.
    """
    history_restore_request: bool = field(init=False)
    """`True` if the request is for history restoration after a miss in the local history cache. Detected by checking if the `HX-History-Restore-Request` header equals `true`."""
    prompt: str | None = field(init=False)
    """The user response to `hx-prompt` if it was used, or `None`."""
    target: str | None = field(init=False)
    """The `id` of the target element if it exists, or `None`. Based on the `HX-Target` header."""
    trigger: str | None = field(init=False)
    """The `id` of the triggered element if it exists, or `None`. Based on the `HX-Trigger` header."""
    trigger_name: str | None = field(init=False)
    """The name of the triggered element if it exists, or `None`. Based on the `HX-Trigger-Name` header."""

    # TODO this requires an HTMX extension, evaluate if it makes sense to use it
    triggering_event: Any = field(init=False)

    def __post_init__(self) -> None:
        self.is_hx_request = self.headers.get(HX_REQUEST) == "true"
        self.boosted = self.headers.get(HX_BOOSTED) == "true"

        # Multi-line logic moved to helpers:
        self.current_url = self._get_current_url()
        self.current_url_abs_path = self._compute_current_url_abs_path(self.current_url)
        self.triggering_event = self._parse_triggering_event(self.headers)

        # Single-line field sets (stay inline):
        self.history_restore_request = self.headers.get(HX_HISTORY_RESTORE_REQUEST) == "true"
        self.prompt = self.headers.get(HX_PROMPT)
        self.target = self.headers.get(HX_TARGET)
        self.trigger = self.headers.get(HX_TRIGGER)
        self.trigger_name = self.headers.get(HX_TRIGGER_NAME)

    def __bool__(self) -> bool:
        """`True` if the request was made with htmx, otherwise `False`. Detected by checking if the `HX-Request` header equals `true`.

        This method allows you to change content for requests made with htmx:

        Example:

            import air
            from random import randint

            app = air.Air()


            @app.page
            def index(request: air.Request):

                if request.htmx:
                    return air.H1(
                        "Click me: ", randint(1, 100),
                        id="number",
                        hx_get="/",
                        hx_swap="outerHTML"
                    )
                return air.layouts.mvpcss(
                    air.H1(
                        "Click me: ", randint(1, 100),
                        id="number",
                        hx_get="/",
                        hx_swap="outerHTML"
                    )
                )
        """

        return self.headers.get("HX-Request") == "true"

    def __str__(self) -> str:
        return str(self.__bool__())

    # ----------------- Private helpers -----------------

    def _get_current_url(self) -> str | None:
        return self.headers.get(HX_CURRENT_URL)

    def _compute_current_url_abs_path(self, url: str | None) -> str | None:
        if url is None:
            return None
        split = urlsplit(url)
        if split.scheme == self.url.scheme and split.netloc == self.url.netloc:
            return urlunsplit(split._replace(scheme="", netloc=""))
        return None

    def _parse_triggering_event(self, headers: Headers) -> Any:
        for name in TRIGGERING_EVENT_ALIASES:
            raw = headers.get(name)
            if raw is not None:
                try:
                    return json.loads(raw)
                except json.JSONDecodeError:
                    return None
        return None


class AirRequest(Request):
    """A wrapper around `starlette.requests.Request` that includes the `HtmxDetails` object.

    !!! note

        AirRequest is available in Air 0.36.0+
    """

    @property
    def htmx(self) -> HtmxDetails:
        return HtmxDetails(headers=self.headers, url=self.url)



"""Default tags, mostly found within the official HTML specification.

Script and Style tags can be found in the [air.tags.models.special](/reference/air.tags.models.special) page."""

from ..utils import locals_cleanup
from .base import AttributesType, BaseTag, Renderable
from .special import SelfClosingTag


class A(BaseTag):
    """Defines a hyperlink

    Args:
        children: Tags, strings, or other rendered content.
        href: Specifies the URL of the page the link goes to.
        target: Specifies where to open the linked document.
        download: Specifies that the target will be downloaded when a user clicks on the hyperlink.
        rel: Specifies the relationship between the current document and the linked document.
        hreflang: Specifies the language of the linked document.
        type: Specifies the media type of the linked document.
        referrerpolicy: Specifies which referrer information to send with the link.
        media: Specifies what media/device the linked document is optimized for.
        ping: Specifies a space-separated list of URLs to which, when the link is followed, post requests with the body ping will be sent by the browser (in the background). Typically used for tracking.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        href: str | None = None,
        target: str | None = None,
        download: str | None = None,
        rel: str | None = None,
        hreflang: str | None = None,
        type: str | None = None,
        referrerpolicy: str | None = None,
        media: str | None = None,
        ping: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Abbr(BaseTag):
    """Defines an abbreviation or an acronym

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Address(BaseTag):
    """Defines contact information for the author/owner of a document

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Area(SelfClosingTag):
    """Defines an area inside an image map

    Args:
        alt: Specifies an alternate text for an area. Required if the href attribute is present.
        coords: Specifies the coordinates of an area.
        download: Specifies that the target will be downloaded when a user clicks on the hyperlink.
        href: Specifies the URL of the page the link goes to.
        ping: Specifies a space-separated list of URLs to which, when the link is followed, post requests with the body ping will be sent by the browser (in the background). Typically used for tracking.
        referrerpolicy: Specifies which referrer information to send with the link.
        rel: Specifies the relationship between the current document and the linked document.
        shape: Specifies the shape of an area.
        target: Specifies where to open the linked document.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *,
        alt: str | None = None,
        coords: str | None = None,
        download: str | None = None,
        href: str | None = None,
        ping: str | None = None,
        referrerpolicy: str | None = None,
        rel: str | None = None,
        shape: str | None = None,
        target: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class Article(BaseTag):
    """Defines an article

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Aside(BaseTag):
    """Defines content aside from the page content

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Audio(BaseTag):
    """Defines embedded sound content

    Args:
        children: Tags, strings, or other rendered content.
        autoplay: Specifies that the audio will start playing as soon as it is ready.
        controls: Specifies that audio controls should be displayed (such as a play/pause button etc).
        loop: Specifies that the audio will start over again, every time it is finished.
        muted: Specifies that the audio output should be muted.
        preload: Specifies if and how the author thinks the audio should be loaded when the page loads.
        src: Specifies the URL of the audio file.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        autoplay: str | None = None,
        controls: str | None = None,
        loop: str | None = None,
        muted: str | None = None,
        preload: str | None = None,
        src: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class B(BaseTag):
    """Defines bold text

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Base(SelfClosingTag):
    """Specifies the base URL/target for all relative URLs in a document

    Args:
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *,
        href: str | None = None,
        target: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class Bdi(BaseTag):
    """Isolates a part of text that might be formatted in a different direction from other text outside it

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Bdo(BaseTag):
    """Overrides the current text direction

    Args:
        children: Tags, strings, or other rendered content.
        dir: Specifies the text direction of the text inside the <bdo> element.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        dir: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Blockquote(BaseTag):
    """Defines a section that is quoted from another source

    Args:
        children: Tags, strings, or other rendered content.
        cite: Specifies the source of the quotation.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        cite: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Body(BaseTag):
    """Defines the document's body

    Args:
        children: Tags, strings, or other rendered content.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Br(SelfClosingTag):
    """Defines a single line break

    Args:
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class Button(BaseTag):
    """Defines a clickable button

    Args:
        children: Tags, strings, or other rendered content.
        name: Specifies a name for the button.
        type: Specifies the type of button.
        value: Specifies an initial value for the button.
        autofocus: Specifies that a button should automatically get focus when the page loads.
        disabled: Specifies that a button should be disabled.
        form: Specifies which form the button belongs to.
        formaction: Specifies where to send the form-data when a form is submitted. Only for type="submit".
        formenctype: Specifies how the form-data should be encoded before sending it to a server. Only for type="submit".
        formmethod: Specifies how to send the form-data (which HTTP method to use). Only for type="submit".
        formnovalidate: Specifies that the form-data should not be validated on submission. Only for type="submit".
        formtarget: Specifies where to display the response that is received after submitting the form. Only for type="submit".
        popovertarget: Specifies which popover element to invoke.
        popovertargetaction: Specifies what action to perform on the popover element.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        name: str | None = None,
        type: str | None = None,
        value: str | None = None,
        autofocus: str | None = None,
        disabled: str | None = None,
        form: str | None = None,
        formaction: str | None = None,
        formenctype: str | None = None,
        formmethod: str | None = None,
        formnovalidate: str | None = None,
        formtarget: str | None = None,
        popovertarget: str | None = None,
        popovertargetaction: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Canvas(BaseTag):
    """Used to draw graphics, on the fly, via scripting (usually JavaScript)

    Args:
        children: Tags, strings, or other rendered content.
        width: Specifies the width of the canvas.
        height: Specifies the height of the canvas.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        width: str | int | None = None,
        height: str | int | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Caption(BaseTag):
    """Defines a table caption

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Cite(BaseTag):
    """Defines the title of a work

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Code(BaseTag):
    """Defines a piece of computer code

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Col(SelfClosingTag):
    """Specifies column properties for each column within a <colgroup> element

    Args:
        span: Specifies the number of columns a <col> element should span.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *,
        span: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class Colgroup(BaseTag):
    """Specifies a group of one or more columns in a table for formatting

    Args:
        children: Tags, strings, or other rendered content.
        span: Specifies the number of columns a <colgroup> element should span.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        span: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Data(BaseTag):
    """Adds a machine-readable translation of a given content

    Args:
        children: Tags, strings, or other rendered content.
        value: Specifies the machine-readable translation of the content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        value: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Datalist(BaseTag):
    """Specifies a list of pre-defined options for input controls

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Dd(BaseTag):
    """Defines a description/value of a term in a description list

    Args:
        children: Tags, strings, or other rendered content.
        cite: Specifies the source of the quotation.
        datetime: Specifies the date and time of the quotation.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        cite: str | None = None,
        datetime: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Del(BaseTag):
    """Defines text that has been deleted from a document

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Details(BaseTag):
    """Defines additional details that the user can view or hide

    Args:
        children: Tags, strings, or other rendered content.
        open: Specifies that the details should be visible (open) to the user.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        open: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Dfn(BaseTag):
    """Specifies a term that is going to be defined within the content

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Dialog(BaseTag):
    """Defines a dialog box or window

    Args:
        children: Tags, strings, or other rendered content.
        open: Specifies that the dialog box should be visible (open) to the user.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        open: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Div(BaseTag):
    """Defines a section in a document

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Dl(BaseTag):
    """Defines a description list

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Dt(BaseTag):
    """Defines a term/name in a description list

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Em(BaseTag):
    """Defines emphasized text

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Embed(SelfClosingTag):
    """Defines a container for an external application

    Args:
        src: Specifies the address of the external file to embed.
        type: Specifies the media type of the embedded content.
        width: Specifies the width of the embedded content.
        height: Specifies the height of the embedded content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *,
        src: str | None = None,
        type: str | None = None,
        width: str | int | None = None,
        height: str | int | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class Fieldset(BaseTag):
    """Groups related elements in a form

    Args:
        children: Tags, strings, or other rendered content.
        disabled: Specifies that a group of related form elements should be disabled.
        form: Specifies which form the fieldset belongs to.
        name: Specifies a name for the fieldset.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        disabled: str | None = None,
        form: str | None = None,
        name: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Figcaption(BaseTag):
    """Defines a caption for a <figure> element

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Figure(BaseTag):
    """Specifies self-contained content

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Footer(BaseTag):
    """Defines a footer for a document or section

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Form(BaseTag):
    """Defines an HTML form for user input

    Args:
        children: Tags, strings, or other rendered content.
        action: Specifies where to send the form-data when a form is submitted.
        method: Specifies the HTTP method to use when sending form-data.
        accept_charset: Specifies the character encodings that are to be used for the form submission.
        autocomplete: Specifies whether a form should have autocomplete on or off.
        enctype: Specifies how the form-data should be encoded when submitting it to the server.
        name: Specifies the name of the form.
        novalidate: Specifies that the form should not be validated when submitted.
        rel: Specifies the relationship between a linked resource and the current document.
        target: Specifies where to display the response that is received after submitting the form.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        action: str | None = None,
        method: str | None = None,
        accept_charset: str | None = None,
        autocomplete: str | None = None,
        enctype: str | None = None,
        name: str | None = None,
        novalidate: str | None = None,
        rel: str | None = None,
        target: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class H1(BaseTag):
    """H1 header

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class H2(BaseTag):
    """H2 header

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class H3(BaseTag):
    """H3 header

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class H4(BaseTag):
    """H4 header

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class H5(BaseTag):
    """H5 header

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class H6(BaseTag):
    """H6 header

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Head(BaseTag):
    """Contains metadata/information for the document

    Args:
        children: Tags, strings, or other rendered content.
        profile: Specifies the URL of a document that contains a line-break-separated list of links.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        profile: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Header(BaseTag):
    """Defines a header for a document or section

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Hgroup(BaseTag):
    """Defines a header and related content

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Hr(SelfClosingTag):
    """Defines a thematic change in the content

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class I(BaseTag):  # noqa: E742
    """Defines a part of text in an alternate voice or mood

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Iframe(BaseTag):
    """Defines an inline frame

    Args:
        children: Tags, strings, or other rendered content.
        src: Specifies the URL of the page to embed.
        srcdoc: Specifies the HTML content of the page to show in the `<iframe>`.
        width: Specifies the width of an `<iframe>`.
        height: Specifies the height of an `<iframe>`.
        allow: Specifies a feature policy for the `<iframe>`.
        allowfullscreen: Set to true if the `<iframe>` can activate fullscreen mode.
        allowpaymentrequest: Set to true if a cross-origin `<iframe>` should be allowed to invoke the Payment Request API.
        loading: Specifies the loading policy of the `<iframe>`.
        name: Specifies the name of an `<iframe>`.
        referrerpolicy: Specifies which referrer information to send when fetching the iframe's content.
        sandbox: Enables an extra set of restrictions for the content in an `<iframe>`.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        src: str | None = None,
        srcdoc: str | None = None,
        width: str | int | None = None,
        height: str | int | None = None,
        allow: str | None = None,
        allowfullscreen: str | None = None,
        allowpaymentrequest: str | None = None,
        loading: str | None = None,
        name: str | None = None,
        referrerpolicy: str | None = None,
        sandbox: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Img(SelfClosingTag):
    """Defines an image

    Args:
        src: Specifies the path to the image.
        width: Specifies the width of an image.
        height: Specifies the height of an image.
        srcset: Specifies a list of image files to use in different situations.
        alt: Specifies an alternate text for an image.
        crossorigin: Allows images from third-party sites that allow cross-origin access to be used with canvas.
        ismap: Specifies an image as a server-side image map.
        loading: Specifies whether a browser should load an image immediately or to defer loading of off-screen images.
        longdesc: Specifies a URL to a detailed description of an image.
        referrerpolicy: Specifies which referrer information to use when fetching an image.
        sizes: Specifies image sizes for different page layouts.
        usemap: Specifies an image as a client-side image map.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *,
        src: str | None = None,
        width: str | int | None = None,
        height: str | int | None = None,
        srcset: str | None = None,
        alt: str | None = None,
        crossorigin: str | None = None,
        ismap: str | None = None,
        loading: str | None = None,
        longdesc: str | None = None,
        referrerpolicy: str | None = None,
        sizes: str | None = None,
        usemap: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class Input(SelfClosingTag):
    """Defines an input control

    Args:
        name: Specifies the name of an `<input>` element.
        type: Specifies the type `<input>` element to display.
        value: Specifies the value of an `<input>` element.
        readonly: Specifies that an input field is read-only.
        required: Specifies that an input field must be filled out before submitting the form.
        accept: Specifies a filter for what file types the user can pick from the file input dialog box.
        alt: Specifies an alternate text for images.
        autocomplete: Specifies whether an `<input>` element should have autocomplete on or off.
        autofocus: Specifies that an `<input>` element should automatically get focus when the page loads.
        checked: Specifies that an `<input>` element should be pre-selected when the page loads.
        dirname: Specifies that the text direction of the input field will be submitted.
        disabled: Specifies that an `<input>` element should be disabled.
        form: Specifies the form the `<input>` element belongs to.
        formaction: Specifies the URL of the file that will process the input control when the form is submitted.
        formenctype: Specifies how the form-data should be encoded when submitting it to the server.
        formmethod: Defines the HTTP method for sending data to the action URL.
        formnovalidate: Specifies that the form-data should not be validated on submission.
        formtarget: Specifies where to display the response that is received after submitting the form.
        height: Specifies the height of an `<input>` element.
        list: Refers to a <datalist> element that contains pre-defined options for an `<input>` element.
        max: Specifies the maximum value for an `<input>` element.
        maxlength: Specifies the maximum number of characters allowed in an `<input>` element.
        min: Specifies a minimum value for an `<input>` element.
        minlength: Specifies the minimum number of characters required in an `<input>` element.
        multiple: Specifies that a user can enter more than one value in an `<input>` element.
        pattern: Specifies a regular expression that an `<input>` element's value is checked against.
        placeholder: Specifies a short hint that describes the expected value of an `<input>` element.
        popovertarget: Specifies which popover element to invoke.
        popovertargetaction: Specifies what action to perform on the popover element.
        size: Specifies the width, in characters, of an `<input>` element.
        src: Specifies the URL of the image to use as a submit button.
        step: Specifies the legal number intervals for an input field.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *,
        name: str | None = None,
        type: str | None = None,
        value: str | None = None,
        readonly: bool | None = None,
        required: bool | None = None,
        accept: str | None = None,
        alt: str | None = None,
        autocomplete: str | None = None,
        autofocus: bool | None = None,
        checked: bool | None = None,
        dirname: str | None = None,
        disabled: bool | None = None,
        form: str | None = None,
        formaction: str | None = None,
        formenctype: str | None = None,
        formmethod: str | None = None,
        formnovalidate: bool | None = None,
        formtarget: str | None = None,
        height: str | int | None = None,
        list: str | None = None,
        max: str | None = None,
        maxlength: str | None = None,
        min: str | None = None,
        minlength: str | None = None,
        multiple: bool | None = None,
        pattern: str | None = None,
        placeholder: str | None = None,
        popovertarget: str | None = None,
        popovertargetaction: str | None = None,
        size: str | None = None,
        src: str | None = None,
        step: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class Ins(BaseTag):
    """Defines a text that has been inserted into a document

    Args:
        children: Tags, strings, or other rendered content.
        cite: Specifies a URL to a document that explains the reason why the text was inserted/changed.
        datetime: Specifies the date and time when the text was inserted/changed.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        cite: str | None = None,
        datetime: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Kbd(BaseTag):
    """Defines keyboard input

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Label(BaseTag):
    """Defines a label for an `<input>` element

    Args:
        children: Tags, strings, or other rendered content.
        for_: Specifies which form element a label is bound to.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        for_: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Legend(BaseTag):
    """Defines a caption for a `<fieldset>` element

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Li(BaseTag):
    """Defines a list item

    Args:
        children: Tags, strings, or other rendered content.
        value: Only for OL lists, this is the starting number of the list item.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        value: int | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Link(SelfClosingTag):
    """Defines the relationship between a document and an external resource (most used to link to style sheets)

    Args:
        href: Specifies the URL of the linked resource.
        as_: Specifies the relationship between the linked resource and the document.
        blocking: Specifies that the resource should be loaded before the rest of the page.
        crossorigin: Specifies how the element handles cross-origin requests.
        disabled: Specifies that the linked resource is disabled.
        fetchpriority: Specifies the priority for fetching the resource.
        hreflang: Specifies the language of the linked resource.
        imagesizes: Specifies the sizes of the icons for visual media.
        imagesrcset: Specifies the URLs of the icons for visual media.
        integrity: Specifies a cryptographic hash of the resource to ensure its integrity.
        media: Specifies the media type of the linked resource.
        referrerpolicy: Specifies which referrer information to send when fetching the resource.
        rel: Specifies the relationship between the current document and the linked resource.
        sizes: Specifies the size of the linked resource.
        title: Specifies the title of the linked resource.
        type: Specifies the media type of the linked resource.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *,
        href: str | None = None,
        as_: str | None = None,
        blocking: str | None = None,
        crossorigin: str | None = None,
        disabled: str | None = None,
        fetchpriority: str | None = None,
        hreflang: str | None = None,
        imagesizes: str | None = None,
        imagesrcset: str | None = None,
        integrity: str | None = None,
        media: str | None = None,
        referrerpolicy: str | None = None,
        rel: str | None = None,
        sizes: str | None = None,
        title: str | None = None,
        type: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class Main(BaseTag):
    """Specifies the main content of a document

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Map(BaseTag):
    """Defines an image map

    Args:
        children: Tags, strings, or other rendered content.
        name: Specifies the name of the image map.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        name: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Mark(BaseTag):
    """Defines marked/highlighted text

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Menu(BaseTag):
    """Defines a menu list

    Args:
        children: Tags, strings, or other rendered content.
        compact: Specifies that the list should be displayed in a compact style.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        compact: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Meta(SelfClosingTag):
    """Defines metadata about an HTML document

    Args:
        charset: Specifies the character encoding for the HTML document.
        content: Specifies the value associated with the http-equiv or name attribute.
        http_equiv: Provides an HTTP header for the information/value of the content attribute.
        media: Specifies what media/device the linked document is optimized for.
        name: Specifies a name for the metadata.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *,
        charset: str | None = None,
        content: str | None = None,
        http_equiv: str | None = None,
        media: str | None = None,
        name: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class Meter(BaseTag):
    """Defines a scalar measurement within a known range (a gauge)

    Args:
        children: Tags, strings, or other rendered content.
        value: The current numeric value. Must be between the min and max values.
        min: The lower bound of the measured range.
        max: The upper bound of the measured range.
        low: The upper numeric bound of the low end of the measured range.
        high: The lower numeric bound of the high end of the measured range.
        optimum: The optimal numeric value.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        value: str | None = None,
        min: str | None = None,
        max: str | None = None,
        low: str | None = None,
        high: str | None = None,
        optimum: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Nav(BaseTag):
    """Defines navigation links

    Args:
         children: Tags, strings, or other rendered content.
         class_: Substituted as the DOM `class` attribute.
         id: DOM ID attribute.
         style: Inline style attribute.
         kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Noscript(BaseTag):
    """Defines an alternate content for users that do not support client-side scripts

    Args:
     children: Tags, strings, or other rendered content.
     class_: Substituted as the DOM `class` attribute.
     id: DOM ID attribute.
     kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Object(BaseTag):
    """Defines an embedded object

    Args:
        children: Tags, strings, or other rendered content.
        archive: A space-separated list of URIs for archives of resources for the object.
        border: The width of a border around the object.
        classidcodebase: The codebase URL for the object.
        codetype: The content type of the code.
        data: The address of the object's data.
        declare: Declares the object without instantiating it.
        form: The form the object belongs to.
        height: The height of the object.
        name: The name of the object.
        standby: A message to display while the object is loading.
        type: The content type of the data.
        usemap: The name of a client-side image map to be used with the object.
        width: The width of the object.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        archive: str | None = None,
        border: str | None = None,
        classid: str | None = None,
        codebase: str | None = None,
        codetype: str | None = None,
        data: str | None = None,
        declare: str | None = None,
        form: str | None = None,
        height: str | int | None = None,
        name: str | None = None,
        standby: str | None = None,
        type: str | None = None,
        usemap: str | None = None,
        width: str | int | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Ol(BaseTag):
    """Defines an ordered list

    Args:
        children: Tags, strings, or other rendered content.
        compact: Specifies that the list should be rendered in a compact style.
        reversed: Specifies that the list order should be descending.
        start: Specifies the start value of an ordered list.
        type: Specifies the kind of marker to use in the list.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        compact: str | None = None,
        reversed: str | None = None,
        start: str | None = None,
        type: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Optgroup(BaseTag):
    """Defines a group of related options in a drop-down list

    Args:
        children: Tags, strings, or other rendered content.
        disabled: Indicates if items in the option group are not selectable.
        label: Specifies a label for the group of options.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        disabled: str | None = None,
        label: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Option(BaseTag):
    """Defines an option in a drop-down list

    Args:
        children: Tags, strings, or other rendered content.
        disabled: Indicates if the option is not selectable.
        label: Specifies a label for the option indicating the meaning of the option.
        selected: Specifies that the option should be pre-selected.
        value: Specifies the value to be sent with the form.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        disabled: str | None = None,
        label: str | None = None,
        selected: bool | None = None,
        value: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Output(BaseTag):
    """Defines the result of a calculation

    Args:
        children: Tags, strings, or other rendered content.
        for_: Lists the IDs of the elements that contributed to the calculation.
        form: Associates the output with a form element.
        name: Defines a name for the output element.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        for_: str | None = None,
        form: str | None = None,
        name: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class P(BaseTag):
    """Defines a paragraph

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Param(SelfClosingTag):
    """Defines a parameter for an object

    Args:
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *,
        class_: str | None = None,
        id: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class Picture(BaseTag):
    """Defines a container for multiple image resources

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Pre(BaseTag):
    """Defines preformatted text

    Args:
        children: Tags, strings, or other rendered content.
        width: preferred counf of characters that a line should have
        wrap: hint indicating how overflow must happen
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        width: str | int | None = None,
        wrap: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Progress(BaseTag):
    """Represents the progress of a task

    Args:
        children: Tags, strings, or other rendered content.
        max: The maximum value of the progress bar.
        value: The current value of the progress bar.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        max: str | None = None,
        value: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Q(BaseTag):
    """Defines a short quotation

    Args:
        children: Tags, strings, or other rendered content.
        cite: Specifies a URL to the source of the quotation.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        cite: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Rp(BaseTag):
    """Defines what to show in browsers that do not support ruby annotations

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Rt(BaseTag):
    """Defines an explanation/pronunciation of characters (for East Asian typography)

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Ruby(BaseTag):
    """Defines a ruby annotation (for East Asian typography)

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class S(BaseTag):
    """Defines text that is no longer correct

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Samp(BaseTag):
    """Defines sample output from a computer program

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Search(BaseTag):
    """Defines a search section

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Section(BaseTag):
    """Defines a section in a document

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Select(BaseTag):
    """Defines a drop-down list

    Args:
        children: Tags, strings, or other rendered content.
        autocomplete: Hint for a user agent's autocomplete feature.
        autofocus: Indicate that a form control should have input focus when the page loads.
        disabled: Indicates that the user cannot interact with the control.
        form: Associates the drop-down list with a form element.
        multiple: Indicates that multiple options can be selected at once.
        name: Specifies the name of the drop-down list.
        required: Indicates that an option must be selected before the form can be submitted.
        size: If drop-down list is a scrolling list box, specifies the number of visible options.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        autocomplete: str | None = None,
        autofocus: str | None = None,
        disabled: str | None = None,
        form: str | None = None,
        multiple: str | None = None,
        name: str | None = None,
        required: str | None = None,
        size: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Small(BaseTag):
    """Defines smaller text

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Source(BaseTag):
    """Defines multiple media resources for media elements (<video> and <audio>)

    Args:
        children: Tags, strings, or other rendered content.
        src: Specifies the URL of the media resource.
        type: Specifies the MIME type of the media resource.
        sizes: List of source sizes that describe the final rendered width of the image.
        media: Specifies the media query for the media resource.
        srcset: Specifies a list of one or more image URLs and their descriptors.
        height: Specifies the height of the media resource.
        width: Specifies the width of the media resource.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        src: str | None = None,
        type: str | None = None,
        sizes: str | None = None,
        media: str | None = None,
        srcset: str | None = None,
        height: str | int | None = None,
        width: str | int | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Span(BaseTag):
    """Defines a section in a document

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Strong(BaseTag):
    """Defines important text

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Sub(BaseTag):
    """Defines subscripted text

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Summary(BaseTag):
    """Defines a visible heading for a <details> element

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Sup(BaseTag):
    """Defines superscripted text

    Args:
        children: Tags, strings, or other rendered content.
        class_: Substituted as the DOM `class` attribute.
        id: DOM ID attribute.
        style: Inline style attribute.
        kwargs: Keyword arguments transformed into tag attributes.
    """

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Table(BaseTag):
    """Defines a table"""

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Tbody(BaseTag):
    """Groups the body content in a table"""

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Td(BaseTag):
    """Defines a cell in a table"""

    def __init__(
        self,
        *children: Renderable,
        colspan: str | None = None,
        rowspan: str | None = None,
        headers: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Template(BaseTag):
    """Defines a container for content that should be hidden when the page loads"""

    def __init__(
        self,
        *children: Renderable,
        shadowrootmode: str | None = None,
        shadowrootdelegatesfocus: str | None = None,
        shadowrootclonable: str | None = None,
        shadowrootserializable: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Textarea(BaseTag):
    """Defines a multiline input control (text area)"""

    def __init__(
        self,
        *children: Renderable,
        autocapitalize: str | None = None,
        autocomplete: str | None = None,
        autocorrect: str | None = None,
        autofocus: bool | None = None,
        cols: str | None = None,
        dirname: str | None = None,
        disabled: bool | None = None,
        form: str | None = None,
        maxlength: str | None = None,
        minlength: str | None = None,
        name: str | None = None,
        placeholder: str | None = None,
        readonly: bool | None = None,
        required: bool | None = None,
        rows: str | None = None,
        spellcheck: str | None = None,
        wrap: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Tfoot(BaseTag):
    """Groups the footer content in a table"""

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Th(BaseTag):
    """Defines a header cell in a table"""

    def __init__(
        self,
        *children: Renderable,
        abbr: str | None = None,
        colspan: str | None = None,
        headers: str | None = None,
        rowspan: str | None = None,
        scope: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Thead(BaseTag):
    """Groups the header content in a table"""

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Time(BaseTag):
    """Defines a specific time (or datetime)"""

    def __init__(
        self,
        *children: Renderable,
        datetime: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Title(BaseTag):
    """Defines a title for the document"""

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Tr(BaseTag):
    """Defines a row in a table"""

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Track(SelfClosingTag):
    """Defines text tracks for media elements (<video> and <audio>)"""

    def __init__(
        self,
        *,
        default: str | None = None,
        kind: str | None = None,
        label: str | None = None,
        srclang: str | None = None,
        src: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))


class U(BaseTag):
    """Defines some text that is unarticulated and styled differently from normal text"""

    def __init__(
        self,
        *children: Renderable,
        compact: str | None = None,
        type: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Ul(BaseTag):
    """Defines an unordered list"""

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Var(BaseTag):
    """Defines a variable"""

    def __init__(
        self,
        *children: Renderable,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Video(BaseTag):
    """Defines embedded video content"""

    def __init__(
        self,
        *children: Renderable,
        src: str | None = None,
        autoplay: str | None = None,
        controls: str | None = None,
        controlslist: str | None = None,
        crossorigin: str | None = None,
        disablepictureinpicture: str | None = None,
        disableremoteplayback: str | None = None,
        height: str | int | None = None,
        width: str | int | None = None,
        loop: str | None = None,
        muted: str | None = None,
        playsinline: str | None = None,
        poster: str | None = None,
        preload: str | None = None,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(*children, **kwargs | locals_cleanup(locals()))


class Wbr(SelfClosingTag):
    """Defines a possible line-break"""

    def __init__(
        self,
        *,
        class_: str | None = None,
        id: str | None = None,
        style: str | None = None,
        **kwargs: AttributesType,
    ) -> None:
        super().__init__(**kwargs | locals_cleanup(locals()))



from collections.abc import Callable

from authlib.integrations.starlette_client import OAuth

from ..requests import Request
from ..responses import RedirectResponse
from ..routing import AirRouter


class GitHubOAuthClientFactory:
    def __init__(
        self,
        github_client_id: str,
        github_client_secret: str,
        github_process_callable: Callable,
        github_redirect_uri: str = "http://localhost:8000/account/github/callback",
        login_redirect_to: str = "/",
        scope="user:email",
    ) -> None:
        """Creates an `air.AirRouter` affiliated with the supplied credentials.

        ARGS:
            github_client_id: The GitHub client ID.
            github_client_secret: The GitHub client secret. Do not include this in your repo, use environment variables!
            github_process_callable: A callable (function, class, or method) that takes three arguments, request, token, and client
            login_redirect_to: The path to send the user to once they have authenticated
            scope: What parts of the GitHub API is accessible

        Example:

            import air
            from os import environ

            app = air.Air()
            app.add_middleware(air.SessionMiddleware, secret_key="change-me")

            async def save_github_token(
                request: air.Request, token: dict, client: Any
            ) -> None:
                "Save the GitHub user's login name to an SQL database."
                resp = await client.get('user', token=token)
                profile = resp.json()
                github_login = profile.get('login')

                async_session = await air.ext.sqlmodel.create_async_session()
                async with async_session() as session:
                    # check if access_token is in database
                    stmt = select(User).where(User.github_login==github_login)
                    result = await session.exec(stmt)
                    user = result.one_or_none()
                    if not user:
                        user = User(
                            github_login=github_login, status=UserStatusEnum.active
                        )
                        session.add(user)
                        await session.commit()
                    # Save the token in place to request.session
                    request.session["user"] = dict(
                        github_login=github_login, updated_at=str(datetime.now())
                    )

            github_oauth_client = air.ext.auth.GitHubOAuthClientFactory(
                github_client_id=environ['GITHUB_CLIENT_ID'],
                github_client_secret=environ['GITHUB_CLIENT_SECRET'],
                github_process_callable=save_github_token,
                github_redirect_uri=environ['GITHUB_REDIRECT_URI']
                login_redirect_to='/dashboard',
                scope='read:profile user:email'
            )
            app.include_router(github_oauth_client.router)
        """
        self.github_redirect_uri = github_redirect_uri
        router = AirRouter()
        oauth = OAuth()
        oauth.register(
            name="github",
            client_id=github_client_id,
            client_secret=github_client_secret,
            access_token_url="https://github.com/login/oauth/access_token",
            access_token_params=None,
            authorize_url="https://github.com/login/oauth/authorize",
            authorize_params=None,
            api_base_url="https://api.github.com/",
            client_kwargs={"scope": scope},
        )
        self.client = oauth.create_client("github")

        @router.get("/account/github/login")
        async def github_login(request: Request):
            assert hasattr(request, "session")
            return await self.client.authorize_redirect(request, self.github_redirect_uri)

        @router.get("/account/github/callback")
        async def github_callback(request: Request):
            assert hasattr(request, "session")
            token = await self.client.authorize_access_token(request)

            await github_process_callable(request=request, token=token, client=self.client)

            return RedirectResponse(login_redirect_to)

        self.router = router
